--[[
##############################################
##-
##-  WakeSpams
##-  By Cyrila <Aspire> @ Bloodfeather-EU
##-
##-  Cooldowns For Your Raid
##-
##-
##-   * Copyright (C) 2009  Dan Jacobsen
##-   *
##-   * This program is free software: you can redistribute it and/or modify
##-   * it under the terms of the GNU General Public License as published by
##-   * the Free Software Foundation, either version 3 of the License, or
##-   * (at your option) any later version.
##-   *
##-   * This program is distributed in the hope that it will be useful,
##-   * but WITHOUT ANY WARRANTY; without even the implied warranty of
##-   * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##-   * GNU General Public License for more details.
##-   *
##-   * You should have received a copy of the GNU General Public License
##-   * along with this program.  If not, see <http://www.gnu.org/licenses/>.
##-
##############################################
]]

local WakeSpams = LibStub("AceAddon-3.0"):NewAddon("WakeSpams", "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0")
_G["WakeSpams"] = WakeSpams

local pairs = pairs;
local ipairs = ipairs;
local sin = _G.math.sin;
local floor = _G.math.floor;
local ceil = _G.math.ceil;
local tinsert = tinsert;
local tremove = tremove;
local tsort = _G.table.sort;
local strlen = _G.string.len;
local strsub = _G.string.sub;
local strfind = _G.string.find;
local strformat = _G.string.format;

local GetSpellInfo = _G["GetSpellInfo"];
local GetTime = _G["GetTime"];

WakeSpams.STANDBY = 0
WakeSpams.Interval = 1000
WakeSpams.Msg = ""
WakeSpams.MsgFrame = "SELF"
WakeSpams.EventDB = {}
WakeSpams.TrackingTypes = {}
WakeSpams.Timer = {}
WakeSpams.LastSent = {0, "", ""}
WakeSpams.RaidIcons = {
	[1] = "{star}", -- Yellow 4-point Star
	[2] = "{circle}", -- Orange Circle
	[3] = "{diamond}", -- Purple Diamond
	[4] = "{triangle}", -- Green Triangle
	[5] = "{moon}", -- White Crescent Moon
	[6] = "{square}", -- Blue Square
	[7] = "{cross}", -- Red 'X' Cross
	[8] = "{skull}"  -- White Skull
}

function WakeSpams:OnInitialize()
	-- Init
	self:RegisterChatCommand("wakespams", "SlashHandler");
end

function WakeSpams:OnEnable()
	-- ENABLED
	self:Print(" -- WakeSpams Loaded! you must look, you must tab, you must see... future. -- ");
	
	-- Register spells and create frame
	self:AddDefaultEvents();
	
	-- Register events
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED", "COMBAT_LOG_EVENT_UNFILTERED");
	self:RegisterEvent("RAID_ROSTER_UPDATE", "RAID_ROSTER_UPDATE");
	self:RegisterEvent("PARTY_MEMBERS_CHANGED", "RAID_ROSTER_UPDATE");
	--self:RegisterEvent("PLAYER_ENTERING_WORLD", "RAID_ROSTER_UPDATE");
end

function WakeSpams:OnDisable()
	-- Do something...
end


--[[
##############################################
##- 
##- Add events (temporary)
##- 
##############################################
]]

function WakeSpams:AddDefaultEvents()
	self:AddTrackingType("SPELL_AURA_APPLIED", 1);
	self:AddTrackingType("SPELL_AURA_REMOVED", 1);
	self:AddTrackingType("SPELL_HEAL", 2);
	self:AddTrackingType("SPELL_CAST_START", 3);
	self:AddTrackingType("SPELL_CAST_SUCCESS", 4);
	self:AddTrackingType("SPELL_MISSED", 5);
	self:AddTrackingType("SPELL_INTERRUPT", 6);
	self:AddTrackingType("SPELL_DISPEL_FAILED", 7);
	self:AddTrackingType("SPELL_RESURRECT", 8);
	
	self.Default = {}
	self.Default.AbilityStart = "Activated: #s# #d#s!"
	self.Default.AbilityEnd = "Faded: #s#"
end

function WakeSpams:AddSpellEvent(spellid, dur, msgt, enda, sn, smsg, emsg, tracktype)
	if not tracktype then tracktype = 1; end;
	tinsert(self.EventDB, {spellid, dur, msgt, enda, sn, smsg, emsg, tracktype});
end

function WakeSpams:AddTrackingType(eventName, ID)
	tinsert(self.TrackingTypes, {ID, eventName});
end


--[[
##############################################
##- 
##- Config
##- 
##############################################
]]


function WakeSpams:SlashHandler(input)
	input = self:Explode(" ", string.lower(input))
	if(input[1] == "enable") then
		if(self.STANDBY == 0) then
			self:Print("Combat log monitoring is already enabled.")
		else
			self.STANDBY = 0
			self:Print("Combat log monitoring enabled.")
		end
	elseif(input[1] == "disable") then
		if(self.STANDBY == 1) then
			self:Print("Combat log monitoring is already disabled.")
		else
			self.STANDBY = 1
			self:Print("Combat log monitoring disabled.")
		end
	else
		DEFAULT_CHAT_FRAME:AddMessage("|cFFFFCC00WakeSpams help:")
		DEFAULT_CHAT_FRAME:AddMessage("|cFFCC9900   Enable|r - Starts combat log monitoring")
		DEFAULT_CHAT_FRAME:AddMessage("|cFFCC9900   Disable|r - Stops combat log monitoring")
		DEFAULT_CHAT_FRAME:AddMessage("|cFFCC9900   Help|r - Shows help")
		DEFAULT_CHAT_FRAME:AddMessage("|cFFCC1100   This addon has no config. You can add and edit spell events in the Lua file.|r")
	end
end


--[[
##############################################
##- 
##- Event Handlers
##- 
##############################################
]]

function WakeSpams:RaidIconText(t)
	local ri = ""
	if UnitExists(t) then
		ri = self.RaidIcons[GetRaidTargetIndex(t)].." " or ri
	end
	return ri
end

function WakeSpams:GetUnitID(GUID, name)
	-- Translates GUID to target (target, focus, mouseover, playername)
	if UnitGUID("focus") == GUID then
		return "focus"
	end
	if UnitGUID("mouseover") == GUID then
		return "mouseover"
	end
	if UnitGUID("target") == GUID then
		return "target"
	end
	return name
end

--
-- HANDLE COMBAT LOG EVENTS AND PASS TO REGISTER SPELL OBJECT
--
function WakeSpams:COMBAT_LOG_EVENT_UNFILTERED(_, _, event, sourceGUID, sourceName, _, destGUID, destName, _, spellid, spellname, _, arg1, arg2)
	if(self.STANDBY == 1) then
		return nil
	end
	if(sourceGUID ~= UnitGUID("player")) then return nil end
	if(spellid) then
		for _,v in pairs(self.TrackingTypes) do
			local i = v[1]
			v = v[2]
			if(strfind(event, v)) then
				for _,t in ipairs(self.EventDB) do
					if((spellid == t[1] or t[1] == "ALL") and t[8] == i) then
					--t[10] = name, rank, icon, count, debuffType, duration, expirationTime, unitCaster, isStealable, shouldConsolidate, spellId
					--print(UnitAura(destName, spellname))
							--t[1] = spellid
							--t[2] = dur
							--t[3] = msgt
							--t[4] = enda
							--t[5] = sn
							--t[6] = smsg
							--t[7] = emsg
							--t[8] = tracktype
						if(event ~= "SPELL_AURA_REMOVED") then
							if not t[6] then
								self.Msg = self.Default.AbilityStart
							else
								self.Msg = t[6]
							end
							
							if(self:isNumeric(t[4]) and (t[4] >= 1)) then
								self.Timer[spellid] = {}
								self.Timer[spellid].elapsed = 0
								self.Timer[spellid].timer = self:ScheduleRepeatingTimer("TimerHandle", 1, {spellid, t[2], t[2]-t[4]}) -- duration, duration-time to know when to start announcing
							end
							--if((self:isNumeric(t[2]) and (t[2] > 0)) and t[3] ~= 2 and event == "SPELL_AURA_APPLIED") then
								--if DXEMOD then
								--	DXEMOD:RaidBar(t[2].." "..GetSpellLink(spellid).." "..destName)
								--end
							--end
						else
							if not t[4] or (t[4] < 0) then
								return nil
							end
							if(t[4] >= 1) then
								-- Stop timers
								self:CancelTimer(self.Timer[spellid].timer)
								self.Timer[spellid] = nil
							end
							if not t[7] then
								self.Msg = self.Default.AbilityEnd
							else
								self.Msg = t[7]
							end
						end
						
						if t[5] then
							spellname = t[5]
						end
						if(spellname == "LINK") then
							spellname = GetSpellLink(spellid)
						end
						-- LETS REPLACE STRING OWOEt
						--"Check the #shit# here.","#(.-)%s"
						self.Msg = string.gsub(self.Msg, "#(.-)#", function(a)
							if(a == "s") then
								a = spellname
							elseif(a == "maxhp") then
								a = floor(UnitHealthMax("player"))
							elseif(a == "halfhp") then
								a = floor(UnitHealthMax("player")*0.5)
							elseif(a == "d") then
								a = t[2]
							elseif(a == "amount") then
								a = arg1
							elseif(a == "extra") then
								if(event == "SPELL_MISSED") then
									a = tostring(arg1)
								else
									a = arg2
								end
							elseif(a == "extra:SpellLink" and event ~= "SPELL_MISSED") then
								a = GetSpellLink(arg1)
							elseif(a == "t") then
								if(a == UnitName("player")) then
									a = "me"
								else
									a = GetUnitName(destName, nil) or destName
									a = self:RaidIconText(self:GetUnitID(destGUID, destName))..a
								end
							end
							return a
						end)
						
						if(t[3] == 2) then
							if(destName == UnitName("player")) then
								self:ChatMSG(self.Msg, "SELF")
							else
								self:ChatMSG(self.Msg, "WHISPER", nil, destName);
							end
						else
							self:ChatMSG(self.Msg, self.MsgFrame);
						end
						break
					end
				end
				break
			end
		end
	end
	--[[if(GetAddOnMemoryUsage("WakeSpams") > 400) then
		if(lastcollect ~= nil and lastcollect > GetTime()-300) then
			-- Cannot collect garbage!!
		else
			lastcollect = GetTime()
			self:Print("Forcing garbage collection!");
			collectgarbage()
		end
	end]]
end

function WakeSpams:RAID_ROSTER_UPDATE()
	if(GetNumRaidMembers() > 1) then
		self.MsgFrame = "RAID"
	elseif(GetNumRaidMembers() <= 0 and GetNumPartyMembers() >= 1) then
		self.MsgFrame = "PARTY"
	else
		self.MsgFrame = "SELF"
	end
	--self:ChatMSG("Raid Roster Update ("..GetNumRaidMembers().." raid members and "..GetNumPartyMembers().." party members). Now sending messages to: "..self.MsgFrame, self.MsgFrame)
end

function WakeSpams:TimerHandle(args)
	-- args: 1 = spellid, 2 = d, 3 = ts
	self.Timer[args[1]].elapsed = self.Timer[args[1]].elapsed + 1
	
	if(self.Timer[args[1]].elapsed >= args[3] and self.Timer[args[1]].elapsed < args[2]) then
		self:ChatMSG(""..args[2] - self.Timer[args[1]].elapsed.."", self.MsgFrame)
	end
end

function WakeSpams:ChatMSG(msg, mtype, mlang, marg)
	if(GetTime() < self.LastSent[1]+self.Interval and self.LastSent[2] == msg and self.LastSent[3] == mtype) then -- Temporary chat throttle
		return nil
	end
	if(mtype=="SELF") then
		DEFAULT_CHAT_FRAME:AddMessage(msg)
	else
		SendChatMessage(msg, mtype, mlang, marg)
	end
	self.LastSent = {GetTime(), msg, mtype}
end

function WakeSpams:isNumeric(a)
    return type(tonumber(a)) == "number"
end

function WakeSpams:ApplyConstant(array)
	return setmetatable ({}, 
	{
		__index = array,
		__newindex = function (t, n, v)
			self:Print("Error: attempting to change constant "..tostring(n).." to "..tostring(v))
		end
    })
end

function WakeSpams:Explode(d, str)
	local t, ll
	t={}
	ll=0
	if(#str == 1) then return str end
		while true do
			l=strfind(str, d, ll+1, true)
			if l~=nil then
				tinsert(t, strsub(str, ll, l-1))
				ll=l+1
			else
				tinsert(t, strsub(str,ll))
				break
			end
		end
	return t
end

--[[
function vardump(value, depth, key)
  local linePrefix = ""
  local spaces = ""
  
  if key ~= nil then
    linePrefix = "["..key.."] = "
  end
  
  if depth == nil then
    depth = 0
  else
    depth = depth + 1
    for i=1, depth do spaces = spaces .. "  " end
  end
  
  if type(value) == 'table' then
    mTable = getmetatable(value)
    if mTable == nil then
      print(spaces ..linePrefix.."(table) ")
    else
      print(spaces .."(metatable) ")
        value = mTable
    end		
    for tableKey, tableValue in pairs(value) do
      vardump(tableValue, depth, tableKey)
    end
  elseif type(value)	== 'function' or 
      type(value)	== 'thread' or 
      type(value)	== 'userdata' or
      value		== nil
  then
    print(spaces..tostring(value))
  else
    print(spaces..linePrefix.."("..type(value)..") "..tostring(value))
  end
end
]]